@@ -2,17 +2,18 @@
  * Copyright (C) 2006, Jamie McCracken <jamiemcc@gnome.org>
  * Copyright (C) 2008-2009, Nokia <ivan.frade@nokia.com>
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the
  * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  * Boston, MA  02110-1301, USA.
  */
@@ -39,16 +40,15 @@
 #include <sys/mman.h>
 #endif /* G_OS_WIN32 */
 
-#ifdef HAVE_ENCA
-#include <enca.h>
-#endif
-
 #include <libtracker-common/tracker-common.h>
 
 #include <libtracker-extract/tracker-extract.h>
 
 #include "tracker-albumart.h"
-#include "tracker-dbus.h"
+
+#ifdef FRAME_ENABLE_TRACE
+#warning Frame traces enabled
+#endif /* FRAME_ENABLE_TRACE */
 
 /* We mmap the beginning of the file and read separately the last 128
  * bytes for id3v1 tags. While these are probably cornercases the
@@ -99,6 +99,8 @@ typedef struct {
 	gchar *title3;
 	gint track_number;
 	gint track_count;
+	gint set_number;
+	gint set_count;
 } id3v2tag;
 
 typedef enum {
@@ -113,6 +115,7 @@ typedef enum {
 	ID3V2_TPB,
 	ID3V2_TP1,
 	ID3V2_TP2,
+	ID3V2_TRK,
 	ID3V2_TT1,
 	ID3V2_TT2,
 	ID3V2_TT3,
@@ -141,6 +144,7 @@ typedef enum {
 	ID3V24_TPE2,
 	ID3V24_TPUB,
 	ID3V24_TRCK,
+	ID3V24_TPOS,
 	ID3V24_TYER,
 } id3v24frame;
 
@@ -148,8 +152,6 @@ typedef struct {
 	size_t size;
 	size_t id3v2_size;
 
-	guint32 duration;
-
 	const gchar *title;
 	const gchar *performer;
 	gchar *performer_uri;
@@ -168,6 +170,8 @@ typedef struct {
 	gchar *composer_uri;
 	gint track_number;
 	gint track_count;
+	gint set_number;
+	gint set_count;
 
 	unsigned char *albumart_data;
 	size_t albumart_size;
@@ -219,6 +223,7 @@ static const struct {
 	{ "TOLY", ID3V24_TOLY },
 	{ "TPE1", ID3V24_TPE1 },
 	{ "TPE2", ID3V24_TPE2 },
+	{ "TPOS", ID3V24_TPOS },
 	{ "TPUB", ID3V24_TPUB },
 	{ "TRCK", ID3V24_TRCK },
 	{ "TYER", ID3V24_TYER },
@@ -239,6 +244,7 @@ static const struct {
 	{ "TP1", ID3V2_TP1 },
 	{ "TP2", ID3V2_TP2 },
 	{ "TPB", ID3V2_TPB },
+	{ "TRK", ID3V2_TRK },
 	{ "TT1", ID3V2_TT1 },
 	{ "TT2", ID3V2_TT2 },
 	{ "TT3", ID3V2_TT3 },
@@ -405,7 +411,7 @@ static const guint freq_mask = 0xC0000;
 static const guint ch_mask = 0xC0000000;
 static const guint pad_mask = 0x20000;
 
-static guint bitrate_table[16][6] = {
+static gint bitrate_table[16][6] = {
 	{   0,   0,   0,   0,   0,   0 },
 	{  32,  32,  32,  32,   8,   8 },
 	{  64,  48,  40,  48,  16,  16 },
@@ -427,7 +433,8 @@ static guint bitrate_table[16][6] = {
 static gint freq_table[4][3] = {
 	{ 44100, 22050, 11025 },
 	{ 48000, 24000, 12000 },
-	{ 32000, 16000, 8000  }
+	{ 32000, 16000,  8000 },
+	{    -1,     -1,   -1 }
 };
 
 static gint spf_table[6] = {
@@ -522,10 +529,10 @@ read_id3v1_buffer (int     fd,
 static gchar *
 ucs2_to_utf8(const gchar *data, guint len)
 {
-	const gchar   *encoding = NULL;
+	const gchar *encoding = NULL;
 	guint16  c;
 	gboolean be;
-	gchar   *utf8 = NULL;
+	gchar *utf8 = NULL;
 
 	memcpy (&c, data, 2);
 
@@ -632,46 +639,24 @@ un_unsync (const unsigned char *source,
 	*dest_size = new_size;
 }
 
-static char*
-get_encoding (const char *data,
-              gssize      size,
-              gboolean   *encoding_found)
+static gchar *
+get_encoding (const gchar *data,
+              gsize        size,
+              gboolean    *encoding_found)
 {
-	gchar *encoding = NULL;
-#ifdef HAVE_ENCA
-	const char **langs;
-	size_t s, i;
-#endif
+	gchar *encoding;
 
+	/* Try to guess encoding */
+	encoding = (data && size ?
+	            tracker_encoding_guess (data, size) :
+	            NULL);
+
+	/* Notify if a proper detection was done */
 	if (encoding_found) {
-		*encoding_found = FALSE;
+		*encoding_found = (encoding ? TRUE : FALSE);;
 	}
 
-#ifdef HAVE_ENCA
-	langs = enca_get_languages (&s);
-
-	for (i = 0; i < s && !encoding; i++) {
-		EncaAnalyser analyser;
-		EncaEncoding eencoding;
-
-		analyser = enca_analyser_alloc (langs[i]);
-		eencoding = enca_analyse_const (analyser, data, size);
-
-		if (enca_charset_is_known (eencoding.charset)) {
-			if (encoding_found) {
-				*encoding_found = TRUE;
-			}
-
-			encoding = g_strdup (enca_charset_name (eencoding.charset,
-			                                        ENCA_NAME_STYLE_ICONV));
-		}
-
-		enca_analyser_free (analyser);
-	}
-
-	free (langs);
-#endif
-
+	/* If no proper detection was done, return default */
 	if (!encoding) {
 		/* Use Windows-1252 instead of ISO-8859-1 as the former is a
 		   superset in terms of printable characters and some
@@ -730,10 +715,6 @@ get_id3 (const gchar *data,
          size_t       size,
          id3tag      *id3)
 {
-#ifdef HAVE_ENCA
-	GString *s;
-	gboolean encoding_was_found;
-#endif /* HAVE_ENCA */
 	gchar *encoding, *year;
 	const gchar *pos;
 
@@ -758,23 +739,31 @@ get_id3 (const gchar *data,
 	 * have a better way to collect a bit more data before we let
 	 * enca loose on it for v1.
 	 */
-#ifdef HAVE_ENCA
-	/* Get the encoding for ALL the data we are extracting here */
-	s = g_string_new_len (pos, 30);
-	g_string_append_len (s, pos + 30, 30);
-	g_string_append_len (s, pos + 60, 30);
+	if (tracker_encoding_can_guess ()) {
+		GString *s;
+		gboolean encoding_was_found;
 
-	encoding = get_encoding (s->str, 90, &encoding_was_found);
+		/* Get the encoding for ALL the data we are extracting here */
+		/* This wont work with encodings where a NUL byte may be actually valid,
+		 * like UTF-16 */
+		s = g_string_new_len (pos, strnlen (pos, 30));
+		g_string_append_len (s, pos + 30, strnlen (pos+30, 30));
+		g_string_append_len (s, pos + 60, strnlen (pos+60, 30));
+		g_string_append_len (s, pos + 94, strnlen (pos+94, ((pos+94)[28] != 0) ? 30 : 28));
 
-	if (encoding_was_found) {
-		id3->encoding = encoding;
+		encoding = get_encoding (s->str, s->len, &encoding_was_found);
+
+		if (encoding_was_found) {
+			id3->encoding = g_strdup (encoding);
+		}
+
+		g_string_free (s, TRUE);
+	} else {
+		/* If we cannot guess encoding, don't even try it, just
+		 * use the default one */
+		encoding = get_encoding (NULL, 0, NULL);
 	}
 
-	g_string_free (s, TRUE);
-#else  /* HAVE_ENCA */
-	encoding = get_encoding (NULL, 0, NULL);
-#endif /* HAVE_ENCA */
-
 	id3->title = g_convert (pos, 30, "UTF-8", encoding, NULL, NULL, NULL);
 
 	pos += 30;
@@ -811,14 +800,15 @@ get_id3 (const gchar *data,
 		id3->genre = g_strdup ("");
 	}
 
-#ifndef HAVE_ENCA
 	g_free (encoding);
-#endif /* HAVE_ENCA */
 
 	return TRUE;
 }
 
-
+/*
+ * For the MP3 frame header description, see
+ * http://www.mp3-tech.org/programmer/frame_header.html
+ */
 static gboolean
 mp3_parse_header (const gchar          *data,
                   size_t                size,
@@ -833,11 +823,11 @@ mp3_parse_header (const gchar          *data,
 	gint spfp8 = 0;
 	guint padsize = 0;
 	gint idx_num = 0;
-	guint bitrate = 0;
+	gint bitrate = 0;
 	guint avg_bps = 0;
 	gint vbr_flag = 0;
 	guint length = 0;
-	guint sample_rate = 0;
+	gint sample_rate = 0;
 	guint frame_size;
 	guint frames = 0;
 	size_t pos = 0;
@@ -890,8 +880,10 @@ mp3_parse_header (const gchar          *data,
 	/* We assume mpeg version, layer and channels are constant in frames */
 	do {
 		frames++;
+
 		bitrate = 1000 * bitrate_table[(header & bitrate_mask) >> 20][idx_num];
 
+		/* Skip frame headers with bitrate index '0000' (free) or '1111' (bad) */
 		if (bitrate <= 0) {
 			frames--;
 			return FALSE;
@@ -899,7 +891,13 @@ mp3_parse_header (const gchar          *data,
 
 		sample_rate = freq_table[(header & freq_mask) >> 18][mpeg_ver - 1];
 
-		frame_size = spfp8 * bitrate / (sample_rate ? sample_rate : 1) + padsize*((header & pad_mask) >> 17);
+		/* Skip frame headers with frequency index '11' (reserved) */
+		if (sample_rate <= 0) {
+			frames--;
+			return FALSE;
+		}
+
+		frame_size = spfp8 * bitrate / sample_rate + padsize*((header & pad_mask) >> 17);
 		avg_bps += bitrate / 1000;
 
 		pos += frame_size;
@@ -943,19 +941,18 @@ mp3_parse_header (const gchar          *data,
 
 	avg_bps /= frames;
 
-	if (filedata->duration == 0) {
-		if ((!vbr_flag && frames > VBR_THRESHOLD) || (frames > MAX_FRAMES_SCAN)) {
-			/* If not all frames scanned
-			 * Note that bitrate is always > 0, checked before */
-			length = (filedata->size - filedata->id3v2_size) / (avg_bps ? avg_bps : bitrate) / 125;
-		} else{
-			length = spfp8 * 8 * frames / (sample_rate ? sample_rate : 0xFFFFFFFF);
-		}
-
-		tracker_sparql_builder_predicate (metadata, "nfo:duration");
-		tracker_sparql_builder_object_int64 (metadata, length);
+	if ((!vbr_flag && frames > VBR_THRESHOLD) || (frames > MAX_FRAMES_SCAN)) {
+		/* If not all frames scanned
+		 * Note that bitrate is always > 0, checked before */
+		length = (filedata->size - filedata->id3v2_size) / (avg_bps ? avg_bps : bitrate) / 125;
+	} else {
+		/* Note that sample_rate is always > 0, checked before */
+		length = spfp8 * 8 * frames / sample_rate;
 	}
 
+	tracker_sparql_builder_predicate (metadata, "nfo:duration");
+	tracker_sparql_builder_object_int64 (metadata, length);
+
 	tracker_sparql_builder_predicate (metadata, "nfo:sampleRate");
 	tracker_sparql_builder_object_int64 (metadata, sample_rate);
 	tracker_sparql_builder_predicate (metadata, "nfo:averageBitrate");
@@ -1102,11 +1099,11 @@ id3v2_text_to_utf8 (const gchar  encoding,
 		                            "Windows-1252",
 		                            NULL, NULL, NULL);
 	case 0x01 :
-		/*              return g_convert (text, */
-		/*                                len, */
-		/*                                "UTF-8", */
-		/*                                "UCS-2", */
-		/*                                NULL, NULL, NULL); */
+		/* return g_convert (text, */
+		/*                   len, */
+		/*                   "UTF-8", */
+		/*                   "UCS-2", */
+		/*                   NULL, NULL, NULL); */
 		return ucs2_to_utf8 (text, len - len%2);
 
 	default:
@@ -1136,10 +1133,10 @@ id3v24_get_frame (const gchar *name)
 	do {
 		m = (l + r) / 2;
 		if (strncmp (name, id3v24_frames[m].name, 4) < 0) {
-			// left half
+			/* left half */
 			r = m - 1;
 		} else {
-			// right half
+			/* right half */
 			l = m + 1;
 		}
 	} while (l <= r && strncmp (id3v24_frames[m].name, name, 4) != 0);
@@ -1164,16 +1161,16 @@ id3v2_get_frame (const gchar *name)
 
 	do {
 		m = (l + r) / 2;
-		if (strncmp (name, id3v2_frames[m].name, 4) < 0) {
-			// left half
+		if (strncmp (name, id3v2_frames[m].name, 3) < 0) {
+			/* left half */
 			r = m - 1;
 		} else {
-			// right half
+			/* right half */
 			l = m + 1;
 		}
-	} while (l <= r && strncmp (id3v2_frames[m].name, name, 4) != 0);
+	} while (l <= r && strncmp (id3v2_frames[m].name, name, 3) != 0);
 
-	if (strncmp (id3v2_frames[m].name, name, 4) == 0) {
+	if (strncmp (id3v2_frames[m].name, name, 3) == 0) {
 		return id3v2_frames[m].frame;
 	} else {
 		return ID3V2_UNKNOWN;
@@ -1258,7 +1255,9 @@ get_id3v24_tags (id3v24frame           frame,
 			g_strstrip (word);
 		}
 
+#ifdef FRAME_ENABLE_TRACE
 		g_debug ("Frame is %d, word is %s", frame, word);
+#endif /* FRAME_ENABLE_TRACE */
 
 		switch (frame) {
 		case ID3V24_TALB:
@@ -1338,6 +1337,21 @@ get_id3v24_tags (id3v24frame           frame,
 
 			break;
 		}
+		case ID3V24_TPOS: {
+			gchar **parts;
+
+			parts = g_strsplit (word, "/", 2);
+			if (parts[0]) {
+				tag->set_number = atoi (parts[0]);
+				if (parts[1]) {
+					tag->set_count = atoi (parts[1]);
+				}
+			}
+			g_strfreev (parts);
+			g_free (word);
+
+			break;
+		}
 		case ID3V24_TYER:
 			if (atoi (word) > 0) {
 				tag->recording_time = tracker_date_guess (word);
@@ -1461,6 +1475,9 @@ get_id3v23_tags (id3v24frame           frame,
 		case ID3V24_TOLY:
 			tag->toly = word;
 			break;
+		case ID3V24_TCOM:
+			tag->composer = word;
+			break;
 		case ID3V24_TIT1:
 			tag->title1 = word;
 			break;
@@ -1498,6 +1515,21 @@ get_id3v23_tags (id3v24frame           frame,
 
 			break;
 		}
+		case ID3V24_TPOS: {
+			gchar **parts;
+
+			parts = g_strsplit (word, "/", 2);
+			if (parts[0]) {
+				tag->set_number = atoi (parts[0]);
+				if (parts[1]) {
+					tag->set_count = atoi (parts[1]);
+				}
+			}
+			g_strfreev (parts);
+			g_free (word);
+
+			break;
+		}
 		case ID3V24_TYER:
 			if (atoi (word) > 0) {
 				tag->recording_time = tracker_date_guess (word);
@@ -1596,6 +1628,21 @@ get_id3v20_tags (id3v2frame            frame,
 		case ID3V2_TP2:
 			tag->performer2 = word;
 			break;
+		case ID3V2_TRK: {
+			gchar **parts;
+
+			parts = g_strsplit (word, "/", 2);
+			if (parts[0]) {
+				tag->track_number = atoi (parts[0]);
+				if (parts[1]) {
+					tag->track_count = atoi (parts[1]);
+				}
+			}
+			g_strfreev (parts);
+			g_free (word);
+
+			break;
+		}
 		case ID3V2_TT1:
 			tag->title1 = word;
 			break;
@@ -1785,10 +1832,10 @@ parse_id3v23 (const gchar          *data,
 		                   ((unsigned char)(data[12]) << 8) |
 		                   ((unsigned char)(data[12]) << 0));
 
-		padding         = (((unsigned char)(data[15]) << 24) |
-		                   ((unsigned char)(data[16]) << 16) |
-		                   ((unsigned char)(data[17]) << 8) |
-		                   ((unsigned char)(data[18]) << 0));
+		padding = (((unsigned char)(data[15]) << 24) |
+		           ((unsigned char)(data[16]) << 16) |
+		           ((unsigned char)(data[17]) << 8) |
+		           ((unsigned char)(data[18]) << 0));
 
 		pos += 4 + ext_header_size;
 
@@ -2111,8 +2158,8 @@ extract_mp3 (const gchar          *uri,
 	                                     md.id3v1.comment);
 
 	md.encoded_by = tracker_coalesce_strip (3, md.id3v24.encoded_by,
-						md.id3v23.encoded_by,
-						md.id3v22.encoded_by);
+	                                        md.id3v23.encoded_by,
+	                                        md.id3v22.encoded_by);
 
 	if (md.id3v24.track_number != 0) {
 		md.track_number = md.id3v24.track_number;
@@ -2132,6 +2179,22 @@ extract_mp3 (const gchar          *uri,
 		md.track_count = md.id3v22.track_count;
 	}
 
+	if (md.id3v24.set_number != 0) {
+		md.set_number = md.id3v24.set_number;
+	} else if (md.id3v23.set_number != 0) {
+		md.set_number = md.id3v23.set_number;
+	} else if (md.id3v22.set_number != 0) {
+		md.set_number = md.id3v22.set_number;
+	}
+
+	if (md.id3v24.set_count != 0) {
+		md.set_count = md.id3v24.set_count;
+	} else if (md.id3v23.set_count != 0) {
+		md.set_count = md.id3v23.set_count;
+	} else if (md.id3v22.set_count != 0) {
+		md.set_count = md.id3v22.set_count;
+	}
+
 	if (md.performer) {
 		md.performer_uri = tracker_sparql_escape_uri_printf ("urn:artist:%s", md.performer);
 
@@ -2186,6 +2249,11 @@ extract_mp3 (const gchar          *uri,
 		tracker_sparql_builder_predicate (preupdate, "nmm:albumTitle");
 		tracker_sparql_builder_object_unvalidated (preupdate, md.album);
 
+		if (md.performer_uri) {
+			tracker_sparql_builder_predicate (preupdate, "nmm:albumArtist");
+			tracker_sparql_builder_object_iri (preupdate, md.performer_uri);
+		}
+
 		tracker_sparql_builder_insert_close (preupdate);
 
 		if (md.track_count > 0) {
@@ -2214,11 +2282,10 @@ extract_mp3 (const gchar          *uri,
 	tracker_sparql_builder_object (metadata, "nmm:MusicPiece");
 	tracker_sparql_builder_object (metadata, "nfo:Audio");
 
-	if (md.title) {
-		tracker_sparql_builder_predicate (metadata, "nie:title");
-		tracker_sparql_builder_object_unvalidated (metadata, md.title);
-	}
-
+	tracker_guarantee_title_from_file (metadata,
+	                                   "nie:title",
+	                                   md.title,
+	                                   uri);
 
 	if (md.lyricist_uri) {
 		tracker_sparql_builder_predicate (metadata, "nmm:lyricist");
@@ -2241,7 +2308,6 @@ extract_mp3 (const gchar          *uri,
 	if (md.album_uri) {
 		tracker_sparql_builder_predicate (metadata, "nmm:musicAlbum");
 		tracker_sparql_builder_object_iri (metadata, md.album_uri);
-		g_free (md.album_uri);
 	}
 
 	if (md.recording_time) {
@@ -2284,6 +2350,53 @@ extract_mp3 (const gchar          *uri,
 		tracker_sparql_builder_object_int64 (metadata, md.track_number);
 	}
 
+	if (md.album) {
+		gchar *album_disc_uri;
+
+		album_disc_uri = tracker_sparql_escape_uri_printf ("urn:album-disc:%s:Disc%d",
+		                                                   md.album,
+		                                                   md.set_number > 0 ? md.set_number : 1);
+
+		tracker_sparql_builder_delete_open (preupdate, NULL);
+		tracker_sparql_builder_subject_iri (preupdate, album_disc_uri);
+		tracker_sparql_builder_predicate (preupdate, "nmm:setNumber");
+		tracker_sparql_builder_object_variable (preupdate, "unknown");
+		tracker_sparql_builder_delete_close (preupdate);
+		tracker_sparql_builder_where_open (preupdate);
+		tracker_sparql_builder_subject_iri (preupdate, album_disc_uri);
+		tracker_sparql_builder_predicate (preupdate, "nmm:setNumber");
+		tracker_sparql_builder_object_variable (preupdate, "unknown");
+		tracker_sparql_builder_where_close (preupdate);
+
+		tracker_sparql_builder_delete_open (preupdate, NULL);
+		tracker_sparql_builder_subject_iri (preupdate, album_disc_uri);
+		tracker_sparql_builder_predicate (preupdate, "nmm:albumDiscAlbum");
+		tracker_sparql_builder_object_variable (preupdate, "unknown");
+		tracker_sparql_builder_delete_close (preupdate);
+		tracker_sparql_builder_where_open (preupdate);
+		tracker_sparql_builder_subject_iri (preupdate, album_disc_uri);
+		tracker_sparql_builder_predicate (preupdate, "nmm:albumDiscAlbum");
+		tracker_sparql_builder_object_variable (preupdate, "unknown");
+		tracker_sparql_builder_where_close (preupdate);
+
+		tracker_sparql_builder_insert_open (preupdate, NULL);
+		tracker_sparql_builder_subject_iri (preupdate, album_disc_uri);
+		tracker_sparql_builder_predicate (preupdate, "a");
+		tracker_sparql_builder_object (preupdate, "nmm:MusicAlbumDisc");
+		tracker_sparql_builder_predicate (preupdate, "nmm:setNumber");
+		tracker_sparql_builder_object_int64 (preupdate, md.set_number > 0 ? md.set_number : 1);
+		tracker_sparql_builder_predicate (preupdate, "nmm:albumDiscAlbum");
+		tracker_sparql_builder_object_iri (preupdate, md.album_uri);
+		tracker_sparql_builder_insert_close (preupdate);
+
+		tracker_sparql_builder_predicate (metadata, "nmm:musicAlbumDisc");
+		tracker_sparql_builder_object_iri (metadata, album_disc_uri);
+
+		g_free (album_disc_uri);
+	}
+
+	g_free (md.album_uri);
+
 	/* FIXME We use a hardcoded value here for now. In reality there's a second option MP3X */
 	tracker_sparql_builder_predicate (metadata, "nmm:dlnaProfile");
 	tracker_sparql_builder_object_string (metadata, "MP3");
@@ -2298,6 +2411,7 @@ extract_mp3 (const gchar          *uri,
 	                          md.album,
 	                          filename);
 	g_free (md.albumart_data);
+	g_free (md.albumart_mime);
 
 	id3v2tag_free (&md.id3v22);
 	id3v2tag_free (&md.id3v23);
